/* BBStree.c
**
** `Balanced `Binary `Search `trees
**     Funktionen zur Verwaltung eines binren Baumes:
**
**     int BBStree_empty (BBStree_t tree)
**         - testen, ob der Baum leer ist
**
**     void BBStree_clear (BBStree_t *tree, void (*elfree) (void *elem))
**         - leeren des Baumes
**
**     void **BBStree_insert (BBStree_t *tree, void *elem,
**                            int (*compare) (void *left, void *right))
**         - einfgen eines neuen Knotens in den Baum; `compare' ist die
**           fr Vergleiche innerhalb der Baum-Struktur bentigte Funktion.
**           Das Ergebnis ist:
**              NULL
**                  fr erfolgreiches Einfgen,
**              ein Zeiger auf einen Zeiger auf das Datenelement
**                  falls schon ein entsprechendes Element vorhanden war
**
**     void **BBStree_get (BBStree_t tree, void *key,
**                         int (*compare) (void *key, void *elem))
**         - lesen eines Knotens aus dem Baumes. Hierbei ist `compare'
**           eine Funktion fr Vergleiche eines Schlssels mit den Elementen
**           des Baum.
**           Das Ergebnis ist:
**              NULL
**                  falls ein Knoten mit dem gegebenen Schlssel nicht
**                  existiert
**              ein Zeiger auf die Speicherstelle in der Baumstruktur, die den
**              Zeiger auf die Daten enthlt
**                  in jedem anderen Fall.
**
**     int BBStree_lotraverse (BBStree_t tree,
**                             int (*eval) (void *node_data, void *static_data),
**                             void *static_data)
**         - Traversieren des gesamten Baumes (pre-order traversal)
**
**     int BBStree_rotraverse (BBStree_t tree,
**                             int (*eval) (void *node_data, void *static_data),
**                             void *static_data)
**         - Traversieren des gesamten Baumes (post-order traversal)
**
**     int BBStree_iotraverse (BBStree_t tree,
**                             int (*eval) (void *node_data, void *static_data),
**                             void *static_data)
**         - Traversieren des gesamten Baumes (in-order traversal)
**
**     Diese Baum-Struktur soll nur ber diese drei Funktionen von auen
**     ansprechbar sein.
*/

#include <stdlib.h>

#include "mrmacs.h"
#include "tools.h"

#define _BBSTREE_C
#include "BBStree.h"

#define BB_ALPHA (0.292893)
#define BB_d (0.414214)

/* Der Implementation beinhaltet die Verwendung von balancierten binren Bumen.
** `weight' dient hierbei zum balancieren des Baumes.
*/
struct bbs_node {
    int weight;
    BBStree_t left;
    BBStree_t right;
    void *data;
};

/* BBStree_empty
**     - berprft, ob ein `tree' leer ist.
*/
int BBStree_empty (BBStree_t tree)
{
    return isNULL(tree);
}

/* BBStree_clear
**     - leert einen `tree'.
*/
void BBStree_clear (BBStree_t *tree, release_func elfree)
{
    ifnull (*tree) return;
    BBStree_clear (&(*tree)->left, elfree);
    BBStree_clear (&(*tree)->right, elfree);
    elfree ((*tree)->data);
    gfree (*tree);
    *tree = (BBStree_t) 0;
}

/* wt
**     berechnet das Gewicht eines Teilbaumes. Dieses ist 1, falls der
**     Teilbaum leer ist und sonst sein gespeichertes Gewicht.
*/
static __inline__ int wt (BBStree_t cc)
{
    return isNULL(cc) ? 1 : cc->weight;
}

/* lrot
**     rotiert die Wurzel und den Linken Unterbaum um eins nach links.
*/
static __inline__ void lrot (BBStree_t *cc)
{
    BBStree_t tmp = *cc;
    *cc = (*cc)->right;
    tmp->right = (*cc)->left;
    (*cc)->left = tmp;
    tmp->weight = wt(tmp->left) + wt(tmp->right);
    (*cc)->weight = wt((*cc)->left) + wt((*cc)->right);
}

/* rrot
**     rotiert die Wurzel und den Linken Unterbaum um eins nach rechts.
*/
static __inline__ void rrot (BBStree_t *cc)
{
    BBStree_t tmp = *cc;
    *cc = (*cc)->left;
    tmp->left = (*cc)->right;
    (*cc)->right = tmp;
    tmp->weight = wt(tmp->left) + wt(tmp->right);
    (*cc)->weight = wt((*cc)->left) + wt((*cc)->right);
}

/* BBStree_insert
**     fgt einen neuen Datensatz `elem' in den Baum ein.
**     Die Funktion `elcomp' dient zur Suche der passenden Stelle zum
**     einfgen.
*/
void **BBStree_insert (BBStree_t *tree, void *elem, compare_func elcomp)
{
    int compval;
    void **rres;
    float bal, bal1;
    ifnull (*tree) {
	/* Einfgen eines Knotens in einen leeren Baum */
	*tree = (BBStree_t) galloc(__FILE__, __LINE__, sizeof(struct bbs_node));
	(*tree)->left   = (BBStree_t) 0;
	(*tree)->right  = (BBStree_t) 0;
	(*tree)->data   = elem;
	(*tree)->weight = 2;
	return NULL;
    }
    /* Daten fr den neuen Knoten mit den Daten des aktuellen Knotens
    ** vergleichen.
    */
    compval = elcomp (elem, (*tree)->data);
    if (!compval) {
	/* Wenn das Element bereits existiert, dann soll es gar nicht mehr
	** eingefgt werden.
	*/
	return &((*tree)->data);
    }
    if (compval < 0) {
	/* Wenn der `Vergleich' ein "kleiner" liefert, dann soll der Knoten in
	** den linken Teilbaum eingefgt werden.
	*/
	rres = BBStree_insert (&(*tree)->left, elem, elcomp);
    } else  {
	/* Wenn der `Vergleich' ein "grer" liefert, dann soll der Knoten in
	** den rechten Teilbaum eingefgt werden.
	*/
	rres = BBStree_insert (&(*tree)->right, elem, elcomp);
    }
    /* Wenn nichts eingefgt wurde (Ergebnis != 0), dann braucht der Baum auch
    ** nicht re-balanciert zu werden.
    */
    unlessnull (rres) { return rres; }
    /* Berechne zunchst die Gewichte neu. */
    (*tree)->weight = wt ((*tree)->left) + wt ((*tree)->right);
    bal = (float) wt ((*tree)->left) / (float) wt (*tree);
    if (bal > 1 - BB_ALPHA) {
	/* Wenn die Fehl-Balance einen bestimmten Wert bersteigt, dann ist der
	** linke Teilbaum offensichtlich zu schwer; balanciere in diesem Fall
	** den Baum durch entweder eine Rechtsrotation des aktuellen Knotens
	** oder durch eine Linksrotation des linken Unterknotens, gefolgt von
	** einer Rechtsrotation des aktuellen Knotens neu.
	*/
	bal1 = (float) wt((*tree)->left->left) / (float) wt((*tree)->left);
	if (bal1 > BB_d) {
	    rrot (tree);
	} else {
	    lrot (&(*tree)->left);
	    rrot (tree);
	}
    } else if (bal < BB_ALPHA) {
	/* Wenn die Fehl-Balance unter einen bestimmten Wert sinkt, dann ist der
	** rechte Teilbaum offensichtlich zu schwer; balanciere in diesem Fall
	** den Baum durch entweder eine Linksrotation des aktuellen Knotens
	** oder durch eine Rechtsrotation des rechten Unterknotens, gefolgt von
	** einer Linksrotation des aktuellen Knotens neu.
	*/
	bal1 = (float) wt((*tree)->right->left) / (float) wt((*tree)->right);
	if (bal1 < 1 - BB_d) {
	    lrot (tree);
	} else {
	    rrot (&(*tree)->right);
	    lrot (tree);
	}
    }
    return (void **) NULL;
}

/* Entferne den Knoten, der zu dem angegebenen Schlssel pat, aus dem Baum.
*/
int BBStree_delete (BBStree_t *tree, void *key, keycomp_func keycomp,
		    release_func elfree)
{
    int compval;
    int rres = 0;
    ifnull (*tree) return 1;
    compval = keycomp (key, (*tree)->data);
    if (compval < 0) {
	rres = BBStree_delete (&(*tree)->left, key, keycomp, elfree);
    } else if (compval > 0) {
	rres = BBStree_delete (&(*tree)->right, key, keycomp, elfree);
    } else ifnull ((*tree)->left) {
	BBStree_t tmp = *tree;
	*tree = (*tree)->right;
	elfree (tmp->data);
	gfree (tmp);
    } else ifnull ((*tree)->right) {
	BBStree_t tmp = *tree;
	*tree = (*tree)->left;
	elfree (tmp->data);
	gfree (tmp);
    } else if (wt((*tree)->left) > wt((*tree)->right)) {
	/* Die linke Seite ist schwerer, darum eine Rechtsrotation */
	if (wt((*tree)->left->left) < wt((*tree)->left->right)) {
	    lrot (&(*tree)->left);
	}
	rrot (tree);
    } else {
	/* Die rechte Seite ist zu schwer, darum eine Linksrotation */
	if (wt((*tree)->right->left) > wt((*tree)->right->left)) {
	    rrot (&(*tree)->right);
	}
	lrot (tree);
    }
    unlessnull (*tree) {
	(*tree)->weight = wt((*tree)->left) + wt((*tree)->right);
    }
    return 0;
}

/* BBStree_get
**     sucht einen Datensatz anhand seines Schlssels `key' innerhalb des
**     Baumes. Wird ein Datensatz passend zu dem gegebenen Schlssel gefunden,
**     dann wird ein Zeiger auf diesen zurckgegeben.
**     Findet BBStree_get hingegen keinen passenden Datensatz, dann ist das
**     Ergebnis NULL.
*/
void **BBStree_get (BBStree_t tree, void *key, keycomp_func keycmp)
{
    int kcres;

    ifnull (tree) return (void **) NULL;

    kcres = keycmp (key, tree->data);
    if (kcres < 0) { return BBStree_get (tree->left, key, keycmp); }
    if (kcres > 0) { return BBStree_get (tree->right, key, keycmp); }
    return &(tree->data);
}

/* Bearbeiten des Baumes in `pre-order' (zunchst der Knoten, dann der linke
** und anschlieend der rechte Teilbaum).
** Hierbei ist das Argument `eval die Funktion, die auf den Daten des
** jeweiligen Knotens ausgefhrt werden soll.
*/
int BBStree_lotraverse (BBStree_t tree, trav_func eval, void *static_data)
{
    int rres;
    unlessnull (tree) {
	if ((rres = eval (tree->data, static_data))) { return rres; }
	if ((rres = BBStree_lotraverse (tree->left, eval, static_data))) {
	    return rres;
	}
	if ((rres = BBStree_lotraverse (tree->right, eval, static_data))) {
	    return rres;
	}
    }
    return 0;
}

/* Bearbeiten des Baumes in `post-order' (zunchst der linke Teilbaum, dann der
** rechte Teilbaum und zum Schlu der aktuelle Knoten).
** Hierbei ist das Argument `eval die Funktion, die auf den Daten des
** jeweiligen Knotens ausgefhrt werden soll.
*/
int BBStree_rotraverse (BBStree_t tree, trav_func eval, void *static_data)
{
    int rres;
    unlessnull (tree) {
	if ((rres = BBStree_lotraverse (tree->left, eval, static_data))) {
	    return rres;
	}
	if ((rres = BBStree_lotraverse (tree->right, eval, static_data))) {
	    return rres;
	}
	if ((rres = eval (tree->data, static_data))) { return rres; }
    }
    return 0;
}

/* Bearbeiten des Baumes in `in-order' (zunchst der linke Teilbaum, dann der
** aktuelle Knoten und anschlieend der rechte Teilbaum).
** Hierbei ist das Argument `eval die Funktion, die auf den Daten des
** jeweiligen Knotens ausgefhrt werden soll.
*/
int BBStree_iotraverse (BBStree_t tree, trav_func eval, void *static_data)
{
    int rres;
    unlessnull (tree) {
	if ((rres = BBStree_lotraverse (tree->left, eval, static_data))) {
	    return rres;
	}
	if ((rres = eval (tree->data, static_data))) { return rres; }
	if ((rres = BBStree_lotraverse (tree->right, eval, static_data))) {
	    return rres;
	}
    }
    return 0;
}
